/*
 * This file and everything in this package is covered under
 * the OPENdj Software License.  See the LICENSE file at the top
 * level of this package for full details.  A copy of the license
 * is also available at: http://opendj.org/docs/licenses/LICENSE.opendj
 * 
 * Copyright 2001 DAX Interactive, LLC.
 *
 * $Author: jonathan $
 * $Revision: 1.1.1.1 $
 * $Header: /home/cvspublic/opendj/opendj-chat-client/src/org/opendj/chat/ChatSocket.java,v 1.1.1.1 2001/09/29 20:20:20 jonathan Exp $
 * $Log: ChatSocket.java,v $
 * Revision 1.1.1.1  2001/09/29 20:20:20  jonathan
 * Imported opendj-chat-client
 *
 */
package org.opendj.chat;

import java.net.*;
import java.io.*;
import java.util.*;

public class ChatSocket implements Runnable, ChatSocketListener {

    // CONSTANTS
    public static final int SOCKDATABUFFER_SIZE = 4096; 
    public static final String SOCKDATA_DELIM     = "^";
    public static final String SOCKDATA_DELIM_ESC = "$__CT__$";
    public static final String SOCKET_SOCKLISTENER_NAME = "------SOCKET------";
    public static final int STATE_ERROR      = -1;
    public static final int STATE_INIT       = 0;
    public static final int STATE_CONNECTING = 1;
    public static final int STATE_CONNECTED  = 2;
    public static final int STATE_INROOM     = 3;
    public static final int STATE_DEAD       = 4;

    // Local info
    private ChatApplet itsApp = null;
    private String itsHost = null;
    private int itsPort = -1;
    private String itsRoomName = null;
    private String itsDJEmail = null;
    private String itsDJPassword = null;

    // Various socket variables
    private Socket itsReadSocket = null;
    private Socket itsWriteSocket = null;
    private DataInputStream itsSockInStream = null;
    private DataOutputStream itsSockOutStream = null;
    private byte[] itsSockDataBuffer = new byte[SOCKDATABUFFER_SIZE];

    private volatile int itsState = STATE_INIT;
    private volatile Thread itsMainThread = null;

    // Everyone listening on the socket
    private Vector itsListeners = null;

    public ChatSocket ( ChatApplet theApp,
                        String theHost, 
                        int thePort, 
                        String theRoomName,
                        String theDJEmail,
                        String theDJPassword ) {
        itsApp = theApp;
        itsHost = theHost;
        itsPort = thePort;
        itsRoomName = theRoomName;
        itsDJEmail = theDJEmail;
        itsDJPassword = theDJPassword;
        itsListeners = new Vector();

        // The socket listens to itself (used to respond to PING commands)
        addListener(this);
    }
    
    public synchronized void joinRoom () {
        // Send hello message to notify server that we've joined the chat
        try {
            sendData(ChatSocketCommands.REQ_HELLO, itsRoomName);
        } catch ( Exception e ) {
            logError("Error sending 'hello' to server to join room", e);
        }
        itsState = STATE_INROOM;
    }

    public synchronized void addListener ( ChatSocketListener theListener ) {

        int aNumListeners = itsListeners.size();
        int aPriority = theListener.getPriority();
        ChatSocketListener aListener = null;

        for ( int i=0; i<aNumListeners; i++ ) {
            aListener = (ChatSocketListener) itsListeners.elementAt(i);
            if ( aPriority <= aListener.getPriority() ) {
                itsListeners.insertElementAt(theListener, i);
                return;
            }
        }

        // Add it to the end
        itsListeners.addElement(theListener);
    }

    public synchronized void removeListener ( ChatSocketListener theListener ) {
        itsListeners.removeElement(theListener);
    }

    public void connect() {
        itsState = STATE_CONNECTING;

        try {
            // Setup socket
            debug("ChatSocket: connecting...");
            itsReadSocket = new Socket(itsHost, itsPort);

            // Setup in stream
            itsSockInStream = new DataInputStream(itsReadSocket.getInputStream());

            // First thing we should receive is a CONFIG that tells us
            // what the out stream port should connect to
            String aConfig = itsSockInStream.readLine();
            StringTokenizer st = new StringTokenizer(aConfig, SOCKDATA_DELIM);
            if ( st.nextToken().equals(ChatSocketCommands.RES_CONFIG) ) {
                int aPort = Integer.parseInt(st.nextToken());
                itsWriteSocket = new Socket(itsHost, aPort);
            } else {
                throw new Exception("Unexpected server response: '" + aConfig + "'");
            }
            itsSockOutStream = new DataOutputStream(itsWriteSocket.getOutputStream());

        } catch ( Exception e ) {
            itsState = STATE_ERROR;
            debug("ChatSocket: error connecting.", e);
            cleanup();
            return;
        }

        itsState = STATE_CONNECTED;
    }

    public void run () {

        itsMainThread = Thread.currentThread();

        // Sanity check
        if ( itsState != STATE_CONNECTED ) return;

        joinRoom();

        if ( itsDJEmail != null ) {
            AutoLoginListener anALListener = new AutoLoginListener(itsApp);
            autoLogin(itsDJEmail, itsDJPassword);
        }


        String anInputLine = null;
        String aCommand = null;
        String aSource = null;
        String aData = null;
        StringTokenizer aTokenizer = null;
        ChatSocketListener aTargetListener = null;
        int aNullReadCount = 0;

        try {
            while ( itsState == STATE_CONNECTED || itsState == STATE_INROOM ) {
                try {
                    try { Thread.sleep(50); } catch ( InterruptedException ie ) {}

                    // Reset vars
                    aCommand = null; aSource = null; aData = null;

                    anInputLine = itsSockInStream.readLine();
                    debug("<" + anInputLine);

                    if ( anInputLine == null ) {
                        if ( ++aNullReadCount > 2 ) {
                            break;
                        }
                        continue;
                    }
                    aNullReadCount = 0;
                    
                    aTokenizer = new StringTokenizer(anInputLine, SOCKDATA_DELIM);
                    
                    if ( aTokenizer.hasMoreTokens() ) {
                        
                        aCommand = aTokenizer.nextToken();
                        if ( aTokenizer.hasMoreTokens() ) {
                            aData = ChatApplet.stringReplace(aTokenizer.nextToken(), 
                                                             SOCKDATA_DELIM_ESC,
                                                             SOCKDATA_DELIM);
                        }
                        if ( aTokenizer.hasMoreTokens() ) {
                            aSource = aTokenizer.nextToken();
                        }
                        
                        // All other messages are delegated to a listener
                        // until someone handles it...
                        processCommand( aCommand, aData, aSource );
                    }
                    
                } catch ( Exception e ) {
                    logError("Error in listen-loop", e);
                }
            }
        } finally {
            cleanup();
        }
    }

    public void shutdown() {
        // Send END to server
        if ( itsState == STATE_CONNECTED || itsState == STATE_INROOM ) {
            sendData(ChatSocketCommands.REQ_END);
        }

        // Remove all listeners
        if ( itsListeners != null ) {
            itsListeners.removeAllElements();
        }

        cleanup();

        itsState = STATE_DEAD;

        if ( itsMainThread != null ) {
            itsMainThread.interrupt();
        }
    }

    public synchronized void reset () {
        try {
            itsReadSocket.close();
        } catch ( Exception e ) {
            logError("ChatSocket.resetUser: Error closing socket", e);
        }
        connect();
        if ( itsMainThread != null ) {
            itsMainThread.interrupt();
        }
        sendData(ChatSocketCommands.REQ_HELLO, itsRoomName);
    }

    public synchronized void cleanup() {
        // Close sockets
        if ( itsReadSocket != null ) {
            try {
                itsReadSocket.close();
            } catch ( Exception e ) {
                logError("Error closing read-socket", e);
            }
        }
        if ( itsWriteSocket != null ) {
            try {
                itsWriteSocket.close();
            } catch ( Exception e ) {
                logError("Error closing write-socket", e);
            }
        }
    }

    public synchronized void sendData ( String theCommand ) {
        sendData( theCommand, null, null );
    }

    public synchronized void sendData ( String theCommand, 
                                        String theMessage ) {
        sendData( theCommand, theMessage, null );
    }

    public synchronized void sendData ( String theCommand,
                                        String theMessage,
                                        String theTarget ) {
        if ( theMessage == null ) theMessage = "";
        if ( theTarget  == null ) theTarget  = "";
        theMessage = ChatApplet.stringReplace( theMessage, SOCKDATA_DELIM, SOCKDATA_DELIM_ESC );

        String someData
            = theCommand
            + SOCKDATA_DELIM
            + theMessage
            + SOCKDATA_DELIM
            + theTarget
            ;

        debug("ChatSocket.sendData: sending '" + someData + "'");
        try {
            sendUnbufferedWrite(itsSockOutStream, someData);
        } catch ( Exception e ) {
            logError("ChatSocket.sendData, error sending '" + someData + "'", e);
        }
    }

    private void sendUnbufferedWrite(DataOutputStream theOutStream, 
                                     String theString) throws IOException {
        theString += ChatApplet.NEWLINE;
        int aStringLen = theString.length();
        
        // Sanity check ... we may have to truncate here...
        if ( aStringLen > SOCKDATABUFFER_SIZE ) {
            theString = theString.substring(0, SOCKDATABUFFER_SIZE-1);
            aStringLen = theString.length();
        }
        
        theString.getBytes(0, aStringLen, itsSockDataBuffer, 0);
        itsSockDataBuffer[aStringLen] = 0;
        theOutStream.write(itsSockDataBuffer, 0, aStringLen);
        theOutStream.flush();
        debug(">" + theString);
    }

    private void logError ( String msg, Exception e ) {
        if ( msg != null ) {
            System.err.println(msg + ": " + e.toString());
        }
        e.printStackTrace();
    }

    private void debug ( Object o ) {
        if ( o == null ) o = new String("null");
        itsApp.debug("[ChatSocket] " + o.toString());
    }

    private void debug( Object o, Exception e ) {
        debug(o);
        e.printStackTrace();
    }

    private synchronized void processCommand ( String theCommand, 
                                               String theData,
                                               String theSource ) {
        ChatSocketListener aListener = null;
        int aNumListeners = itsListeners.size();

        for ( int i=0; i<aNumListeners; i++ ) {
            aListener = (ChatSocketListener) itsListeners.elementAt(i);

            // If the listener handled the command, stop processing
            try {
                if ( aListener.handleServerCommand( theCommand, 
                                                    theData, 
                                                    theSource ) ) {
                    return;
                }
            } catch ( Exception e ) {
                logError("Error calling handleServerCommand", e);
            }
        }
        debug("Unhandled server message '" + theCommand + "'");
    }

    public void autoLogin ( String theEmail, String theEPassword ) {
        
        debug("autoLogin called");
        while ( itsState != STATE_INROOM ) {
            debug("autoLogin waiting to join room");
            try { Thread.sleep(150); } catch ( Exception e ) {}
        }
        debug("sending auto login data to server");
        sendData( ChatSocketCommands.REQ_JOINASDJEP, theEmail, theEPassword );
    }

    public void finalize () {
        shutdown();
    }

    // ChatSocketListener interface methods
    public String getName() { return SOCKET_SOCKLISTENER_NAME; }

    // Socket listener always gets first crack at commands
    public int getPriority() { return Integer.MAX_VALUE; }

    public boolean handleServerCommand( String theCommand, 
                                        String theSource, 
                                        String theData ) {
        if (theCommand.equals(ChatSocketCommands.RES_PING)) {
            // Respond to ping with pong
            sendData(ChatSocketCommands.REQ_PONG);
            return true;
        }
        return false;
    }
}
